/*
*  mmlp_grid.c
*  Minimax label propagation for superpixel segmentation on grid
*
*  Author: Kye-Hyeon Kim, Hyeonseob Nam
*  Contact: {fenrir,namhs09}@postech.ac.kr
*  Last updated: January, 2016
*
*  Programmed and tested on Windows 7 64-bit and MATLAB R2012a
*
*  COMPILE
*  -----------------------------------------------------------------------
*  mex -v OPTIMFLAGS="/Ox" -largeArrayDims mmlp_grid.c
* 
*  USAGE
*  -----------------------------------------------------------------------
*  [F, V, P, K_final] = mmlp_grid(seedset, C_right, C_down, F_init, V_init, P_init, LABEL_init)
*
*  F: height-by-width matrix
*  F(i,j): 1,2,...,C  the predicted superpixel label for the (i,j) pixel
*
*  INPUTS 
*
*  V: height-by-width matrix
*  V(i,j): the minimax distance from its seed pixel 
*
*  P: height-by-width matrix
*  P(i,j): the index of its parent
*
*  K_final: the final number of labels
*
*  OUTPUTS
*
*  seedset: N_L-by-1 vector,  i: 0,...,N_L-1
*  seedset(i): 1,...,N  the actual node index of the i-th seed pixel
*
*  C_right: height-by-(width-1) matrix
*  C_right(u,v): the distance between pixel (u,v) and its right pixel (u,v+1)
*
*  C_down: height-by-width matrix
*  C_down(u,v): the distance between pixel (u,v) and its lower pixel (u+1,v)
*               the bottom row C_down(height,:) is ignored
*
*  F_init: height-by-width matrix
*  F(i,j): 1,2,...,C  the inital superpixel label for the (i,j) pixel
*
*  V: height-by-width matrix
*  V(i,j): the initial minimax distance from its seed pixel 
*
*  P: height-by-width matrix
*  P(i,j): the initial index of its parent
*
*  LABEL_init: the number of initial labels
*
*/

#include "mex.h"
#include <string.h>
#include <math.h>

// input arguments
// (the class labels and minimax distances are directly updated and stored
//  in the arrays f and v, respectively)
#define NUM_ARG_IN		7
#define ARY_SEED		prhs[0]
#define MAT_COST_RIGHT	prhs[1]
#define MAT_COST_DOWN	prhs[2]
#define MAT_F_INIT		prhs[3]
#define MAT_V_INIT		prhs[4]
#define MAT_P_INIT		prhs[5]
#define LABEL_INIT		prhs[6]
//#define VAL_REGULARITY  prhs[7]
//#define MERGE_SMALLSEG	prhs[8]

// output arguments
#define NUM_ARG_OUT		4
#define ARY_F			plhs[0]
#define ARY_V			plhs[1]
#define ARY_P			plhs[2]
#define K_FINAL			plhs[3]

void mexFunction( int nlhs, mxArray *plhs[],
	int nrhs, const mxArray *prhs[] )
{
	/*********************** VARIABLE DECLARATION ***********************/
	unsigned int width, height;  // width and height of the image
	unsigned int N;  // the number of pixels
	unsigned int N_L;  // the number of seed pixels

	unsigned int *seedset;  // vector of indices of seed nodes
	unsigned int *f;  // vector of predicted superpixel labels
	unsigned int *v;  // vector of the corresponding minimax distances
	unsigned int *f_out;
	unsigned int *v_out;
	unsigned int *p_out;
	unsigned int k_final;

	unsigned int *C_right;  // link cost to right adjacent pixel, height-by-(width-1) matrix
	unsigned int *C_down;  // link cost to lower adjacent pixel, (height-1)-by-width matrix
	bool *is_top;  // indicator whether the i-th pixel is at the top row of the image
	bool *is_bottom;  // indicator whether the i-th pixel is at the top bottom of the image

	// variables for breath-first search during the label propagation process
	unsigned int *Q;  // a circular queue of nodes waiting for propagation
	unsigned int Q_max_length, Q_first, Q_last;  // max, first, last indices in Q
	bool *is_in_Q;  // to indicate whether each node is in Q

	// variables for ensuring that every superpixel is a connected component
	unsigned int *parent;  // the parent index of the i-th pixel in the minimax path tree
	// if the (i+1)-th pixel is seed, parent[i] = i+1 (itself)
	bool label_changed;  // temporary indicator whether the label of a node is changed

	// variables for assigning a small superpixel to its nearest adjacent superpixel
	// ("nearest" in the notion of minimax distance)
	unsigned int *cluster_size;
	unsigned int small_threshold;
	unsigned int *v_near_min;
	unsigned int *parent_near_min;
	unsigned int *root_near_min;

	// inital labels & minimax distances & parents
	unsigned int *f_init;
	unsigned int *v_init;
	unsigned int *p_init;
	// inital number of labels
	unsigned int label_init;

	// variables for controlling the regularity of size and shape of superpixels
	unsigned int v_penalty;
	// flag for merging small segments
	bool merge_small_seg;

	// indices and temporary variables
	unsigned int i, j, t;
	unsigned int val;

	/*********************** INITIALIZATION ***********************/
	// check command error
	if (nrhs < NUM_ARG_IN)
		mexErrMsgTxt("Incorrect number of input arguments.");
	if (nlhs > NUM_ARG_OUT)
		mexErrMsgTxt("Too many output arguments");

	// get width / height / the number of pixels and seed pixels
	width = (unsigned int)mxGetN(MAT_COST_DOWN);
	height = (unsigned int)mxGetM(MAT_COST_RIGHT);
	N = width*height;
	N_L = (unsigned int)mxGetNumberOfElements(ARY_SEED);

	// memory allocation and initializion
	seedset = (unsigned int *)mxGetData(ARY_SEED);
	f = mxCalloc(N, sizeof *f);
	v = mxMalloc(N*(sizeof *v));

	C_right = (unsigned int *)mxGetData(MAT_COST_RIGHT);
	C_down = (unsigned int *)mxGetData(MAT_COST_DOWN);
	is_top = mxCalloc(N, sizeof *is_top);
	is_bottom = mxCalloc(N, sizeof *is_bottom);

	Q = mxMalloc(N*(sizeof *Q));
	is_in_Q = mxCalloc(N, sizeof *is_in_Q);

	parent = mxCalloc(N, sizeof *parent);

	// mark whether each pixel is either in the top row or the bottom row
	for (i = 0; i < N; i += height) {
		is_top[i] = true;
	}
	for (i = height-1; i < N; i += height) {
		is_bottom[i] = true;
	}

	// initilize labels
	label_init = (unsigned int)mxGetScalar(LABEL_INIT);
	if (mxGetNumberOfElements(MAT_F_INIT) > 0) {
		f_init = (unsigned int *)mxGetData(MAT_F_INIT);
		for (i = 0; i < N; i++) {
			f[i] = f_init[i];
		}
	}

	// initilize distances
	if (mxGetNumberOfElements(MAT_V_INIT) > 0) {
		v_init = (unsigned int *)mxGetData(MAT_V_INIT);
		for (i = 0; i < N; i++)
			v[i] = v_init[i];
	}
	else {
		for (i = 0; i < N; i++)
			v[i] = (unsigned int)(-1); //MAX_UINT
	}

	// initilize parents
	if (mxGetNumberOfElements(MAT_P_INIT) > 0) {
		p_init = (unsigned int *)mxGetData(MAT_P_INIT);
		for (i = 0; i < N; i++)
			parent[i] = p_init[i];
	}

	// push all N_L seed nodes into Q
	// N_L seed nodes are initially inserted into Q as seeds for propagation
	for (t = 0; t < N_L; ++t) {
		i = seedset[t]-1;
		f[i] = label_init+t+1;
		v[i] = 0;
		Q[t] = i;
		is_in_Q[i] = true;
		parent[i] = i;
	}

	//v_penalty = (unsigned int)(1000 * mxGetScalar(VAL_REGULARITY) / sqrt(4.0*N/N_L));
	//v_penalty = (unsigned int)(1000 * mxGetScalar(VAL_REGULARITY));
    v_penalty = 0;
	//merge_small_seg = (bool)mxGetScalar(MERGE_SMALLSEG);
    merge_small_seg = 1;
    
	Q_max_length = N-1;
	Q_first = 0;  // seed nodes are now in Q, from Q[0] to Q[N_L-1]
	Q_last = N_L;

	/*********************** MAIN ALGORITHM ***********************/
	// perform minimax label propagation
	while (1)
	{
		// pop a node (say node i) from Q to propagate the label of the node
		if (Q_first == Q_last) {
			break;
		}

		i = Q[Q_first];
		Q_first = (Q_first+1)%Q_max_length;

		// now node i is not in Q
		is_in_Q[i] = false;

		// visit each adjacent node of node i (say node j)
		// to propagate the label assigned to node i into node j

		// 1. check left adjacent node
		if (i >= height)
		{
			j = i - height;

			val = max(v[i], C_right[j]) + v_penalty;

			if (val < v[j])
			{
				// update the corresponding minimax distance
				// and node j's parent
				v[j] = val;
				f[j] = f[i];
				parent[j] = i;

				// push this updated node j into Q for further propagation
				// unless j is already in Q
				if (!is_in_Q[j]) {
					Q[Q_last] = j;
					Q_last = (Q_last+1)%Q_max_length;
					is_in_Q[j] = true;
				}
			}
		}

		// 2. check right adjacent node
		j = i + height;
		if (j < N)
		{
			val = max(v[i], C_right[i]) + v_penalty;

			if (val < v[j])
			{
				// update the corresponding minimax distance
				// and node j's parent
				v[j] = val;
				f[j] = f[i];
				parent[j] = i;

				// push this updated node j into Q for further propagation
				// unless j is already in Q
				if (!is_in_Q[j]) {
					Q[Q_last] = j;
					Q_last = (Q_last+1)%Q_max_length;
					is_in_Q[j] = true;
				}
			}
		}

		// 3. check upper adjacent node
		if (!is_top[i])
		{
			j = i - 1;

			val = max(v[i], C_down[j]) + v_penalty;

			// if node i's label is updated,
			// all of its children and descendants' labels are also updated
			// (for ensuring that every superpixel is a connected component)
			if (val < v[j])
			{
				// update the corresponding minimax distance
				// and node j's parent
				v[j] = val;
				f[j] = f[i];
				parent[j] = i;

				// push this updated node j into Q for further propagation
				// unless j is already in Q
				if (!is_in_Q[j]) {
					Q[Q_last] = j;
					Q_last = (Q_last+1)%Q_max_length;
					is_in_Q[j] = true;
				}
			}
		}

		// 4. check lower adjacent node
		if (!is_bottom[i])
		{
			j = i + 1;

			val = max(v[i], C_down[i]) + v_penalty;

			// if node i's label is updated,
			// all of its children and descendants' labels are also updated
			// (for ensuring that every superpixel is a connected component)
			if (val < v[j])
			{
				// update the corresponding minimax distance
				// and node j's parent
				v[j] = val;
				f[j] = f[i];
				parent[j] = i;

				// push this updated node j into Q for further propagation
				// unless j is already in Q
				if (!is_in_Q[j]) {
					Q[Q_last] = j;
					Q_last = (Q_last+1)%Q_max_length;
					is_in_Q[j] = true;
				}
			}
		}
	}

	/*********************** CONSTRAINED MMLP ***********************/
	t = 0;
	for (i = 0; i < N; ++i) {
		if (f[i] != f[parent[i]]) {
			Q[t++] = i;
			is_in_Q[i] = true;
		}
	}
	
	Q_first = 0;
	Q_last = t;

	while (1)
	{
		// pop a node (say node i) from Q to propagate the label of the node
		if (Q_first == Q_last) {
			break;
		}

		if (Q_first < Q_max_length) {
			i = Q[Q_first++];
		}
		else {
			Q_first = 0;
			i = Q[Q_max_length];
		}

		// label propagation: node i's parent -> node i
		f[i] = f[parent[i]];

		// now node i is not in Q
		is_in_Q[i] = false;

		// visit each adjacent node of node i (say node j)
		// to propagate the label assigned to node i into node j

		// 1. check left adjacent node
		if (i >= height)
		{
			j = i - height;

			// if node i's label is updated,
			// all of its children and descendants' labels are also updated
			// (for ensuring that every superpixel is a connected component)
			if (i == parent[j])
			{
				// update the corresponding minimax distance
				// (v[j] >= val for all children)
				val = max(v[i], C_right[j]) + v_penalty;
				v[j] = val;

				// push this updated node j into Q for further propagation
				// unless j is already in Q
				if (!is_in_Q[j]) {
					if (Q_last < Q_max_length) {
						Q[Q_last++] = j;
					}
					else {
						Q_last = 0;
						Q[Q_max_length] = j;
					}
					is_in_Q[j] = true;
				}
			}
		}

		// 2. check right adjacent node
		j = i + height;
		if (j < N)
		{
			// if node i's label is updated,
			// all of its children and descendants' labels are also updated
			// (for ensuring that every superpixel is a connected component)
			if (i == parent[j])
			{
				// update the corresponding minimax distance
				// (v[j] >= val for all children)
				val = max(v[i], C_right[i]) + v_penalty;
				v[j] = val;

				// push this updated node j into Q for further propagation
				// unless j is already in Q
				if (!is_in_Q[j]) {
					if (Q_last < Q_max_length) {
						Q[Q_last++] = j;
					}
					else {
						Q_last = 0;
						Q[Q_max_length] = j;
					}
					is_in_Q[j] = true;
				}
			}
		}

		// 3. check upper adjacent node
		if (!is_top[i])
		{
			j = i - 1;

			// if node i's label is updated,
			// all of its children and descendants' labels are also updated
			// (for ensuring that every superpixel is a connected component)
			if (i == parent[j])
			{
				// update the corresponding minimax distance
				// (v[j] >= val for all children)
				val = max(v[i], C_down[j]) + v_penalty;
				v[j] = val;

				// push this updated node j into Q for further propagation
				// unless j is already in Q
				if (!is_in_Q[j]) {
					if (Q_last < Q_max_length) {
						Q[Q_last++] = j;
					}
					else {
						Q_last = 0;
						Q[Q_max_length] = j;
					}
					is_in_Q[j] = true;
				}
			}
		}

		// 4. check lower adjacent node
		if (!is_bottom[i])
		{
			j = i + 1;

			// if node i's label is updated,
			// all of its children and descendants' labels are also updated
			// (for ensuring that every superpixel is a connected component)
			if (i == parent[j])
			{
				// update the corresponding minimax distance
				// (v[j] >= val for all children)
				val = max(v[i], C_down[i]) + v_penalty;
				v[j] = val;

				// push this updated node j into Q for further propagation
				// unless j is already in Q
				if (!is_in_Q[j]) {
					if (Q_last < Q_max_length) {
						Q[Q_last++] = j;
					}
					else {
						Q_last = 0;
						Q[Q_max_length] = j;
					}
					is_in_Q[j] = true;
				}
			}
		}
	}


	/*********************** POST-PROCESSING ***********************/
	N_L = N_L + label_init;
	cluster_size = mxMalloc((N_L+1)*(sizeof *cluster_size));
	v_near_min = mxMalloc((N_L+1)*(sizeof *v_near_min));
	parent_near_min = mxMalloc((N_L+1)*(sizeof *parent_near_min));
	root_near_min = mxMalloc((N_L+1)*(sizeof *root_near_min));

	// count the size of each superpixel
	for (t = 1; t<N_L+1; t++)
		cluster_size[t] = 0;

	for (i = 0; i < N; ++i)
		++cluster_size[f[i]];

	if (merge_small_seg) {

		// set threshold of the minimum size of superpixel to be remained
		// initial threshold = 10% of the average size
		small_threshold = (unsigned int)min(100,((0.1*N) / N_L));
		
		// assign a tiny superpixel to its nearest, larger adjacent superpixel
		// ("nearest" in the notion of minimax distance)

		// for each superpixel, find its nearest, larger adjacent superpixel
		// ("nearest" in the notion of minimax distance)
		for (t = 1; t<N_L+1; t++)
			v_near_min[t] = (unsigned int)(-1); //MAX_UINT

		for (i = 0; i < N; ++i) {
			if (cluster_size[f[i]] >= small_threshold) continue;

			if (i < N-height && f[i] != f[i+height]) {
				val = max(v[i+height], C_right[i]) + v_penalty;
				if (v_near_min[f[i]] > val
					&& cluster_size[f[i+height]] > cluster_size[f[i]]) {
						v_near_min[f[i]] = val;
						parent_near_min[f[i]] = i+height;
						root_near_min[f[i]] = i;
				}
			}

			if (i >= height && f[i] != f[i-height]) {
				val = max(v[i-height], C_right[i-height]) + v_penalty;
				if (v_near_min[f[i]] > val
					&& cluster_size[f[i-height]] > cluster_size[f[i]]) {
						v_near_min[f[i]] = val;
						parent_near_min[f[i]] = i-height;
						root_near_min[f[i]] = i;
				}
			}

			if (!is_top[i] && f[i] != f[i-1]) {
				val = max(v[i-1], C_down[i-1]) + v_penalty;
				if (v_near_min[f[i]] > val
					&& cluster_size[f[i-1]] > cluster_size[f[i]]) {
						v_near_min[f[i]] = val;
						parent_near_min[f[i]] = i-1;
						root_near_min[f[i]] = i;
				}
			}

			if (!is_bottom[i] && f[i] != f[i+1]) {
				val = max(v[i+1], C_down[i]) + v_penalty;
				if (v_near_min[f[i]] > val
					&& cluster_size[f[i+1]] > cluster_size[f[i]]) {
						v_near_min[f[i]] = val;
						parent_near_min[f[i]] = i+1;
						root_near_min[f[i]] = i;
				}
			}
		}

		// assign a tiny superpixel to its nearest, larger adjacent superpixel
		for (t = 1; t <N_L+1; t++) {

			if (cluster_size[t] < small_threshold && v_near_min[t]!=(unsigned int)(-1)) {

				cluster_size[t] = 0;
				i = root_near_min[t];
				parent[i] = parent_near_min[t];
				Q[0] = i;
				is_in_Q[i] = true;

				Q_first = 0;
				Q_last = 1;

				while (1)
				{
					if (Q_first == Q_last) break;

					i = Q[Q_first++];
					f[i] = f[parent[i]];

					if (i >= height && f[i-height] == t) {
						j = i-height;
						parent[j] = i;
						if (!is_in_Q[j]) {
							Q[Q_last++] = j;
							is_in_Q[j] = true;
						}
					}

					if (i < N-height && f[i+height] == t) {
						j = i+height;
						parent[j] = i;
						if (!is_in_Q[j]) {
							Q[Q_last++] = j;
							is_in_Q[j] = true;
						}
					}

					if (!is_top[i] && f[i-1] == t) {
						j = i-1;
						parent[j] = i;
						if (!is_in_Q[j]) {
							Q[Q_last++] = j;
							is_in_Q[j] = true;
						}
					}

					if (!is_bottom[i] && f[i+1] == t) {
						j = i+1;
						parent[j] = i;
						if (!is_in_Q[j]) {
							Q[Q_last++] = j;
							is_in_Q[j] = true;
						}
					}
				}
			}
		}
	}

	k_final = N_L;
	for (t = 1; t<N_L+1; t++) {
		if(cluster_size[t] == 0) k_final--;
	}

	ARY_F = mxCreateNumericMatrix(height, width, mxUINT32_CLASS, mxREAL);
	f_out = (unsigned int *)mxGetData(ARY_F);
	for (i = 0; i < N; ++i) {
		f_out[i] = f[i];
	}

	ARY_V = mxCreateNumericMatrix(height, width, mxUINT32_CLASS, mxREAL);
	v_out = (unsigned int *)mxGetData(ARY_V);
	for (i = 0; i < N; ++i) {
		v_out[i] = v[i];
	}

	ARY_P = mxCreateNumericMatrix(height, width, mxUINT32_CLASS, mxREAL);
	p_out = (unsigned int *)mxGetData(ARY_P);
	for (i = 0; i < N; ++i) {
		p_out[i] = parent[i];
	}

	K_FINAL = mxCreateDoubleScalar(k_final);

	/*************** FINISHING ***************/
	// memory deallocation
	mxFree(f);
	mxFree(v);
	mxFree(is_top);
	mxFree(is_bottom);
	mxFree(Q);
	mxFree(is_in_Q);
	mxFree(parent);
	mxFree(cluster_size);
	mxFree(v_near_min);
	mxFree(parent_near_min);
	mxFree(root_near_min);

	return;
}