function [Y, K, time] = mmlp_superpixel(imgname, K)
% INPUTS
% imgname (string): the path to input image
% k : the desired number of superpixels
% 
% OUTPUTS
% Y (HxW matrix): the output superpixel labels for all pixels
% K : the output number of superpixels
% time : processing time

%% Image loading
img = imread(imgname);

t_start = tic;

X = int32(img);
height = size(X, 1);
width = size(X, 2);

%% Link cost computation on the grid
C_right = uint32(mean(abs(X - X(:,[2:end end],:)),3));
C_down  = uint32(mean(abs(X - X([2:end end],:,:)),3));

%% compute edge map
E = edgemap(img);
E = uint32(E*255);

%% compute edge distance
D_right = 1000*(C_right+(E+E(:,[2:end end]))/2);
D_down = 1000*(C_down+(E+E([2:end end],:))/2);

%% Initial Seed selection
K_cur = K - min(100,floor(K * 0.5));

seed_gap = sqrt(width*height/K_cur);
seed_x = floor(seed_gap/2 + linspace(1, width-seed_gap, round(width/seed_gap)));
seed_y = floor(seed_gap/2 + linspace(1, height-seed_gap, round(height/seed_gap)));
[seed_x seed_y] = meshgrid(seed_x, seed_y);
seed = uint32(sub2ind([height, width], seed_y(:), seed_x(:)));

%% Move seed to flat neighbor
seed_neighbor = [seed-height-1 seed-height seed-height+1 ...
    seed-1 seed seed+1 ...
    seed+height-1 seed+height seed+height+1];

grad_mean = D_right(:,[1 1:end-1]) + D_right + D_down([1 1:end-1],:) + D_down;
grad_seed_neighbor = grad_mean(seed_neighbor);
[~, min_grad_idx] = min(grad_seed_neighbor, [], 2);
seed = seed_neighbor(sub2ind(size(seed_neighbor),(1:length(min_grad_idx))',min_grad_idx));

%% Classification
[Y, minmaxMap, P, K_cur] = mmlp_grid(seed, D_right, D_down, [], [], [], 0);

r = ceil(seed_gap/16);
for i=1:2
    %% New seed selection
    K_add = K-K_cur;
    if (K_add < 0.05*K), break; end
    
    H = fspecial('average',2*r+1);
    minmaxMap_ = single(minmaxMap/1000);
    minmaxFlat = abs(minmaxMap_ - convolve2(minmaxMap_,H,'same')) < 0.001;
    
    L = bwlabel(minmaxFlat);
    L_idx = find(L(:));
    [L_nnz_sort, L_nnz_idx] = sort(L(L_idx));
    IA = L_idx(L_nnz_idx([true;diff(L_nnz_sort)>0]));
    
    [~, id] = sort(minmaxMap(IA),'descend');
    seed_new = uint32(IA(id(1:min(end,K_add))));
    
    r = ceil(r/2);
    
    %% Mode seed to flat neighbor
    [s_i, s_j] = ind2sub([height, width], seed_new);
    s_i = max(2, min(height-1, s_i));
    s_j = max(2, min(width-1, s_j));
    seed_new = sub2ind([height width],s_i,s_j);
    
    seed_neighbor = [seed_new-height-1 seed_new-height seed_new-height+1 ...
        seed_new-1 seed_new seed_new+1 ...
        seed_new+height-1 seed_new+height seed_new+height+1];
    
    grad_seed_neighbor = grad_mean(seed_neighbor);
    [~, min_grad_idx] = min(grad_seed_neighbor, [], 2);
    seed_new = seed_neighbor(sub2ind(size(seed_neighbor),(1:length(min_grad_idx))',min_grad_idx));
    
    %% Classification
    [Y, minmaxMap, P, K_cur] = mmlp_grid(seed_new, D_right, D_down, Y, minmaxMap, P, numel(seed));
    seed = [seed; seed_new];
end

time = toc(t_start);
disp(['    Elapsed time: ' num2str(time)]);

%% Boundary & output
[~, ~, Y] = unique(Y);
K = max(Y);
Y = reshape(Y,height,width);
Y = uint16(Y);

end

